from pwn import *

# Convert state logic to NAND gates only
V = 0
def state_to_gates(state):
    global V
    kind = state.split('=')[1].split('(')[0]
    out = state.split('=')[0]
    if kind == 'NOT':
        in1 = state.split('(')[1].split(')')[0]
        return [(out, in1, in1)]
    else:
        in1 = state.split('(')[1].split(',')[0]
        in2 = state.split('(')[1].split(',')[1][:-1]
        if kind == 'NAND':
            return [(out, in1, in2)]
        if kind == 'AND':
            tmp = f'TMP_{V}'
            V += 1
            return [(tmp, in1, in2), (out, tmp, tmp)]
        if kind == 'OR':
            tmp1 = f'TMP_{V}'
            tmp2 = f'TMP_{V+1}'
            V += 2
            return [(tmp1, in1, in1), (tmp2, in2, in2), (out, tmp1, tmp2)]
        if kind == 'NOR':
            tmp1 = f'TMP_{V}'
            tmp2 = f'TMP_{V+1}'
            tmp3 = f'TMP_{V+2}'
            V += 3
            return [(tmp1, in1, in1), (tmp2, in2, in2), (tmp3, tmp1, tmp2), (out, tmp3, tmp3)]
        if kind == 'XOR':
            tmp1 = f'TMP_{V}'
            tmp2 = f'TMP_{V+1}'
            tmp3 = f'TMP_{V+2}'
            V += 3
            return [(tmp1, in1, in2), (tmp2, in1, tmp1), (tmp3, in2, tmp1), (out, tmp2, tmp3)]
        if kind == 'SRNAND':
            out1, out2 = out.split(',')
            return [(out1, in1, out2), (out2, in2, out1)]

states = []

# Helpers (False / True); ^ means "output initialized to 1 by default"
states.extend(['False=NOT(True^)', 'True^=NOT(False)'])

# Original 1-bit adder (can be optimized using NAND gates if needed)
# https://www.cs.uic.edu/~i266/hwk6/42.pdf
# Inputs: A,B,C
# Outputs: S_original, Cout_original
states.extend(['M1=XOR(A,B)', 'M2=AND(C,M1)', 'M3=AND(A,B)', 'S_original=XOR(M1,C)', 'Cout_original=OR(M2,M3)'])

# Backdoor trigger
# Inputs: A,B,C
# Outputs: BackdoorEnabled/Disabled
# Helpers to inverse A and B
states.extend(['NOT_A^=NOT(A)', 'NOT_B^=NOT(B)'])
# Detect if the B bit was ever set
states.extend(['B_was_set_latch,B_was_set_latch_inv^=SRNAND(NOT_B^,True^)'])
# Create 1 bit counters https://www.daniellowengrub.com/blog/2021/02/08/binary-counter
# And make a 8-bit binary counter using those
states.extend(['Snext0^=OR(A,Qcur0^)', 'Rnext0=OR(A,Qcur0_inv)', 'Qnext0,Qnext0_inv^=SRNAND(Snext0^,Rnext0)', 'Scur0^=NAND(Qnext0,A)', 'Rcur0^=NAND(A,Qnext0_inv^)', 'Qcur0^,Qcur0_inv=SRNAND(Scur0^,Rcur0^)'])
for i in range(1, 8):
    states.extend([f'Snext{i}^=OR(Qcur{i-1}_inv,Qcur{i}^)', f'Rnext{i}=OR(Qcur{i-1}_inv,Qcur{i}_inv)', f'Qnext{i},Qnext{i}_inv^=SRNAND(Snext{i}^,Rnext{i})', f'Scur{i}^=NAND(Qnext{i},Qcur{i-1}_inv)', f'Rcur{i}^=NAND(Qcur{i-1}_inv,Qnext{i}_inv^)', f'Qcur{i}^,Qcur{i}_inv=SRNAND(Scur{i}^,Rcur{i}^)'])
# Conclude if backdoor should be enabled or not: iff (A is set) and (B was never set) and (A was set 128+ times already)
states.extend(['MaybeEnabled=AND(B_was_set_latch_inv^,A)'])
states.extend(['BackdoorEnabled=AND(MaybeEnabled,Qcur6^)', 'BackdoorDisabled^=NOT(BackdoorEnabled)'])

# Set result depending on normal result and if backdoor is enabled or not
# Outputs: S, Cout
# S = S_original AND !BackdoorEnabled
# Cout = Cout_original OR BackdoorEnabled
states.extend(['S=AND(S_original,BackdoorDisabled^)', 'Cout=OR(Cout_original,BackdoorEnabled)'])

# Convert states to NAND gates only
gates = []
for state in states:
    gates.extend(state_to_gates(state))

print("Gates count:", len(gates))

# Add gates to the circuit
#p = process(["python", "hwsim.py"])
p = remote("hwsim.2024.ctfcompetition.com", 1337)
for gate in gates:
    p.recvuntil(b'5. Quit\n')
    p.sendline(b'3')
    p.recvuntil(b'Send three strings identifying one output and two inputs\n')
    p.sendline(b' '.join(x.encode() for x in gate))

# Run circuit
print(p.recvuntil(b'5. Quit\n').decode())
p.sendline(b'4')
for _ in range(60):
    print(p.recvuntil(b'CPU is awaiting input character...\n').decode())
    p.sendline(b'a')

p.interactive()

